###############################################################################
#
# Copyright (c) 2014 Projekt01 GmbH.
# All Rights Reserved.
#
###############################################################################
"""Tests
$Id: tests.py 3934 2014-03-17 07:38:52Z roger.ineichen $
"""
from __future__ import absolute_import
from __future__ import print_function
__docformat__ = "reStructuredText"

import doctest
import re
import unittest
from zope.testing import renormalizing

checker = renormalizing.RENormalizing([
    (re.compile('\r\n'), '\n'),
    ])


def test_suite():
    return unittest.TestSuite((
        doctest.DocFileSuite('checker.txt',
                             globs={'print_function': print_function,
                                    'absolute_import': absolute_import},
                             ),
        doctest.DocFileSuite('../README.txt',
                             optionflags=doctest.NORMALIZE_WHITESPACE | doctest.ELLIPSIS,
                             globs={'print_function': print_function,
                                    'absolute_import': absolute_import},
                             checker=checker
                             ),
        doctest.DocFileSuite('zcml.txt',
                             optionflags=doctest.NORMALIZE_WHITESPACE | doctest.ELLIPSIS,
                             checker=checker
                             ),
    ))

if __name__ == '__main__':
    unittest.main(defaultTest='test_suite')



##############################################################################
#
# Copyright (c) 2006 Zope Foundation and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################
"""Templated Resource Processor

$Id: zrt.py 5278 2025-05-15 14:55:28Z roger.ineichen $
"""
from __future__ import absolute_import

__docformat__='restructuredtext'

import re
import zope.interface
import zope.component
from zope.interface import implementer
from zope.pagetemplate import engine

from p01.cdn import interfaces


################################################################################
#
# expressions

# <EXPR-TYPE>"<INPUT-EXPR>" <EXPR-TYPE>"<OUTPUT-EXPR>" <NUM>
NAME = r'[a-zA-Z0-9_-]*'
ARGS_REGEX = re.compile(r' *(%s)"([^"]*)" *(%s)"([^"]*)" *([0-9]*)' %(NAME, NAME))


class BaseExpression(object):

    def __init__(self, source, context, request):
        self.source = source
        self.context = context
        self.request = request


class StringInputExpression(BaseExpression):
    """A simple string input expression"""
    zope.interface.implements(interfaces.IZRTInputExpression)

    def process(self, text, outputExpr, count=None):
        regex = re.compile(re.escape(self.source))
        return regex.subn(outputExpr.process(), text, count or 0)


class StringOutputExpression(BaseExpression):
    """A simple string input expression"""
    zope.interface.implements(interfaces.IZRTOutputExpression)

    def process(self, **kw):
        # Ignore any keyword arguments, since this is static replacement
        return self.source


class RegexInputExpression(BaseExpression):
    """A regex string input expression"""
    zope.interface.implements(interfaces.IZRTInputExpression)

    def process(self, text, outputExpr, count=None):
        regex = re.compile(self.source)
        return regex.subn(outputExpr.process(), text, count or 0)


class TALESOutputExpression(BaseExpression):
    """A simple string input expression"""
    zope.interface.implements(interfaces.IZRTOutputExpression)

    def process(self, **kw):
        expr = engine.TrustedEngine.compile(self.source)
        kw.update({'context': self.context, 'request': self.request})
        econtext = engine.TrustedEngine.getContext(kw)
        return expr(econtext)


class Replace(object):
    """A ZRT Command to replace sub-strings of the text"""
    zope.interface.implements(interfaces.IZRTCommand)

    inputExpressions = {
        '': StringInputExpression,
        'str': StringInputExpression,
        're': RegexInputExpression,
        }

    outputExpressions = {
        '': StringOutputExpression,
        'str': StringOutputExpression,
        'tal': TALESOutputExpression,
        }

    def __init__(self, args, start, end):
        self.start = start
        self.end = end
        self.processArguments(args)

    @property
    def isAvailable(self):
        return self.num is None or self.num > 0

    def processArguments(self, args):
        match = ARGS_REGEX.match(args)
        if match is None:
            raise interfaces.ArgumentError(args)

        self.itype, self.input, self.otype, self.output, self.num = match.groups()
        self.num = self.num and int(self.num) or None

        if self.itype not in self.inputExpressions:
            raise ValueError(self.itype)

        if self.otype not in self.outputExpressions:
            raise ValueError(self.otype)

    def process(self, text, context, request):
        iexpr = self.inputExpressions[self.itype](
            self.input, context, request)
        oexpr = self.outputExpressions[self.otype](
            self.output, context, request)
        text, num = iexpr.process(text, oexpr, self.num)
        if self.num is not None:
            self.num -= num
        return text


################################################################################
#
# cdn expressions

@implementer(interfaces.IZRTOutputExpression)
class CDNInputExpression(BaseExpression):
    """A simple string input expression"""

# XXX: probably RegexInputExpression is also broken
    def process(self, text, outputExpr, count=None):
        """Process simple string without backslash convertion

        Prevent backslash convertion, see re.sub documenations which says:
        any backslash escapes in it are processed. That is, \n is converted
        to a single newline character.
        This shold fix an issue with json2.js which has the following code:

        meta = {    // table of character substitutions
           '\b': '\\b',
           '\t': '\\t',
           '\n': '\\n',
           '\f': '\\f',
           '\r': '\\r',
           '"' : '\\"',
           '\\': '\\\\'
        },
        """
        regex = re.compile(re.escape(r'%s' % self.source))
        return regex.subn(r'%s' % outputExpr.process(), r'%s' % text, count or 0)


@implementer(interfaces.IZRTOutputExpression)
class CDNIncludeOutputExpression(BaseExpression):
    """CDN resource include expression"""

    def process(self, **kw):
        # Ignore any keyword arguments, since this is static replacement
        name = self.source
        resource = zope.component.queryAdapter(self.request, name=name)
        f = open(resource.path, 'rb')
        data = f.read()
        f.close()
        data = data.decode('utf-8', 'replace')
        # NOTE, this replaces the missing structure concept and prevents from
        # replace '\n' with '' and '\\n' with '\n' see json2.js as a sample
        return data.replace('\\', '\\\\')


@implementer(interfaces.IZRTCommand)
class CDNReplace(Replace):
    """ASame as Replace with additional cdn include expression

    You can use this additional expression as:

    /* zrt-replace:"REPLACE_THIS" include"RESOURCE_NAME" */
    REPLACE_THIS
    """

    inputExpressions = {
        '': CDNInputExpression,
        'str': CDNInputExpression,
        're': RegexInputExpression,
        }

    outputExpressions = {
        '': StringOutputExpression,
        'str': StringOutputExpression,
        'tal': TALESOutputExpression,
        'include': CDNIncludeOutputExpression,
        }



##############################################################################
#
# Copyright (c) 2009 Projekt01 GmbH and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################
"""
$Id:$
"""
from __future__ import absolute_import
from future import standard_library
standard_library.install_aliases()
from builtins import str
__docformat__ = "reStructuredText"

import urllib.request, urllib.parse, urllib.error
import zope.component
import zope.interface
import zope.i18nmessageid
import zope.location.interfaces
from zope.proxy import sameProxiedObjects
from zope.publisher.browser import BrowserView
from zope.traversing.browser.interfaces import IAbsoluteURL
from zope.publisher.interfaces.browser import IBrowserRequest

from p01.cdn import interfaces

_ = zope.i18nmessageid.MessageFactory('p01')

_insufficientContext = _("There isn't enough context to get URL information. "
                         "This is probably due to a bug in setting up location "
                         "information.")

_safe = '@+' # Characters that we don't want to have quoted


class AbsoluteURLForCDNResource(BrowserView):

    zope.interface.implementsOnly(IAbsoluteURL)
    zope.component.adapts(interfaces.ICDNResource, IBrowserRequest)

    def __unicode__(self):
        return urllib.parse.unquote(self.__str__()).decode('utf-8')

    def __str__(self):
        uri = self.context()
        return urllib.parse.quote(uri.encode('utf-8'), _safe)

    __call__ = __str__

    def breadcrumbs(self):
        context = self.context
        request = self.request

        # We do this here do maintain the rule that we must be wrapped
        context = zope.location.interfaces.ILocation(context, context)
        container = getattr(context, '__parent__', None)
        if container is None:
            raise TypeError(_insufficientContext)

        if sameProxiedObjects(context, request.getVirtualHostRoot()) or \
               isinstance(context, Exception):
            return ({'name':'', 'url': self.request.getApplicationURL()}, )

        base = tuple(zope.component.getMultiAdapter(
                (container, request), name='absolute_url').breadcrumbs())

        name = getattr(context, '__name__', None)
        if name is None:
            raise TypeError(_insufficientContext)

        if name:
            base += ({'name': name, 'url': str(self)},)

        return base



##############################################################################
#
# Copyright (c) 2009 Projekt01 GmbH and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################
"""
$Id:$
"""
from __future__ import absolute_import
from zope.interface import implementer
from builtins import object
__docformat__ = "reStructuredText"

import json
import os.path
import posixpath

import zope.interface
from zope.schema.fieldproperty import FieldProperty

from p01.cdn import interfaces


ZRT_START = u"""/*---[ START: zrt-replace.less ]----------------------------------------------*/
"""

ZRT_END = u"""/*---[ END: zrt-replace.less ]------------------------------------------------*/
"""


@implementer(interfaces.IResourceManager)
class ResourceManager(object):
    """Resource Manager"""


    _version = FieldProperty(interfaces.IResourceManager['version'])
    _namespace = FieldProperty(interfaces.IResourceManager['namespace'])
    _skin = FieldProperty(interfaces.IResourceManager['skin'])
    _site = FieldProperty(interfaces.IResourceManager['site'])
    rawURI = FieldProperty(interfaces.IResourceManager['rawURI'])
    _uri = FieldProperty(interfaces.IResourceManager['uri'])
    _output = FieldProperty(interfaces.IResourceManager['output'])
    useSVNVersions = FieldProperty(
        interfaces.IResourceManager['useSVNVersions'])
    svnVersionSourcePath = FieldProperty(
        interfaces.IResourceManager['svnVersionSourcePath'])
    zrtDirReplaces = FieldProperty(interfaces.IResourceManager['zrtDirReplaces'])
    zrtReplaces = FieldProperty(interfaces.IResourceManager['zrtReplaces'])
    svnZRTReplacePath = FieldProperty(
        interfaces.IResourceManager['svnZRTReplacePath'])
    versions = FieldProperty(interfaces.IResourceManager['versions'])

    def __init__(self, version, uri, output=None, namespace=None, skin=None,
        site=None, useSVNVersions=False, svnVersionSourcePath=None,
        svnZRTReplacePath=None):
        # set uri arguments
        self.version = version
        self.namespace = namespace
        self.skin = skin
        self.site = site
        # setup converted uri
        self.uri = uri
        # setup output
        self.output = output
        # subversion and package version based resource versioning
        self.useSVNVersions = useSVNVersions
        self.svnVersionSourcePath = svnVersionSourcePath
        self.svnZRTReplacePath = svnZRTReplacePath
        self.setUpVersions()

    def setUpVersions(self):
        """Load extracted resource versions"""
        self.versions = {}
        self.zrtDirReplaces = u''
        self.zrtReplaces = u''
        if self.svnVersionSourcePath is not None:
            if not os.path.exists(self.svnVersionSourcePath):
                f = open(self.svnVersionSourcePath, 'wb')
                f.close()
            # use version (don't enable during extraction)
            if self.useSVNVersions:
                f = open(self.svnVersionSourcePath, 'rb')
                fData = f.read()
                f.close()
                if fData:
                    self.versions = json.loads(fData)

    def addVersion(self, name, resource, options):
        """Add resource version

        The name is the traversable resource lookup name. The rName is the
        resource name and the fName is the external file name.

        NOTE: this method is only used during resource version setup. This
        optional step is only used if you use file based resource versions.
        The p01.recipe.cdn setup script is using this method for apply the
        version for each file based on svn repository or egg package version
        if the resource is located in an egg package.

        Note: you can simply implement your own version lookup concept if you
        use GIT or another concept. This method is all you need for define and
        setup the right version for the given resource. The p01.recipe.cdn
        extract method will later lookup the version for each file from the
        prepopulated self.versions dict.
        """
        # find svn or package version
        version = options.svn.getRevision(resource)
        # add zrt-repalce directive (p01.recipe.cdn extract will save them)
        zrtReplace = resource.getZRTReplace(name, version, options)
        if zrtReplace is not None:
            if resource.__parent__ is not None:
                self.zrtDirReplaces += zrtReplace
            else:
                self.zrtReplaces += zrtReplace
        # add version (p01.recipe.cdn extract will save them)
        # now, use resource.rName whcih could be different then the given name
        # attribute. Ok, most the time it's the same name but it could be
        # different if you implement some magic directory with a naming pattern
        # like we do with our i18n resources where we generate the rName
        f1, f2 = os.path.splitext(resource.rName)
        fName = '%s.v%s%s' % (f1, version, f2)
        self.versions[resource.rName] = fName
        return fName

    def checkVersions(self):
        """Check changed versions (used before write new files)

        Note: write a new file forces the less compiler to comile less to css.
        We will prevent this in p01.recipe.cdn if we didn't have new versions
        """
        if self.svnVersionSourcePath is not None:
            if os.path.isfile(self.svnVersionSourcePath):
                try:
                    f = open(self.svnVersionSourcePath, 'rb')
                    org = json.loads(f.read())
                    f.close()
                except ValueError:
                    # ValueError: No JSON object could be decoded
                    # empty file or bad format, mark as changed
                    return False
            else:
                org = {}
            if self.versions == org:
                # version not changed
                return True
            else:
                # version changed
                return False
        else:
            # version config not found on file system
            return False

    def saveVersions(self):
        """Store version mapping and zrt-replace.less in file system

        NOTE: this method is only used during resource extraction
        """
        # write cdn.json file
        if self.svnVersionSourcePath is not None:
            data = json.dumps(self.versions, indent=4, sort_keys=True)
            f = open(self.svnVersionSourcePath, 'wb')
            f.write(data)
            f.close()

    def saveZRTReplace(self):
        """Store version mapping and zrt-replace.less in file system

        NOTE: this method is only used during resource extraction
        """
        # write zrt-include.less file
        if self.svnZRTReplacePath is not None:
            data = ZRT_START
            data += self.zrtDirReplaces
            data += self.zrtReplaces
            data += ZRT_END
            f = open(self.svnZRTReplacePath, 'wb')
            f.write(data)
            f.close()

    def _convert(self):
        """Converts previous set uri with dynamic values

        This ensures that we convert the uri during change on everytime we
        access them.
        """
        version = self.version or 'missing-version'
        namespace = self.namespace or 'missing-namespace'
        skin = self.skin or 'missing-skin'
        site = self.site or 'missing-site'
        self._uri = self.rawURI % {'version': version,
            'namespace': namespace, 'skin': skin, 'site': site}

    def version():
        def fget(self):
            return self._version
        def fset(self, value):
            self._version = value
            # convertraw uri to uri
            self._convert()
        return property(fget, fset)

    version = version()
    def namespace():
        def fget(self):
            return self._namespace
        def fset(self, value):
            self._namespace = value
            # convertraw uri to uri
            self._convert()
        return property(fget, fset)

    namespace = namespace()
    def skin():
        def fget(self):
            return self._skin
        def fset(self, value):
            self._skin = value
            # convertraw uri to uri
            self._convert()
        return property(fget, fset)

    skin = skin()
    def site():
        def fget(self):
            return self._site
        def fset(self, value):
            self._site = value
            # convertraw uri to uri
            self._convert()
        return property(fget, fset)

    site = site()
    def uri():
        def fget(self):
            return self._uri
        def fset(self, uri):
            # keep a reference to the uri including formatting strings
            self.rawURI = uri
            self._convert()
        return property(fget, fset)

    uri = uri()
    def output():
        """Return dynamicly converted output path"""
        def fget(self):
            if self._output is not None:
                return self._output % {
                    'version': self.version,
                    'namespace': self.namespace,
                    }
            else:
                return self._output
        def fset(self, output):
            self._output = output
        return property(fget, fset)

    output = output()
    def getExtractFileName(self, name, resource):
        """Returns the relative resource path on extraction"""
        # get name from cdn.json resource version map or use given name
        rName = resource.getRelativePath(name)
        # check version map for correct resource name by relative resource path
        rName = self.versions.get(rName, rName)
        # just return the filename including a version if any
        return rName.split('/')[-1]

    def getURI(self, name=None):
        """Get the correct url based on the uri, namespace and version and
        resource name.

        We also, allow to use * as a version manager marker.

        An initial devmode uri could look like:

        http://localhost:8080/++skin++Admin/%(version)s/@@

        If your applicatiton uses site and subsite and each sub site is using
        another resource manager, the relative subsite path could get used
        as namespace. Then your uri could look like:

        http://localhost:8080/++skin++Admin/%(namespace)s/%(version)s/@@

        And the sub site could be set to something like: "root/subsite"

        A production setup does not require any special uri setup in general.
        It fully depends on your web servers rewrite rule and your extracted
        resource location. You will probably use a sub domain like:

        http://%(namespace)s.foobar.com/%(version)s/@@

        or with an additional namespace:

        http://cdn.foobar.com/%(namespace)s/%(version)s/@@

        or just as minimal as possible

        http://cdn.foobar.com/%(version)s/@@

        Note: you can also use the site and skin as a part of the uri
        formatting.

        """
        if name is None:
            # return base uri
            return self.uri
        else:
            # get name from cdn.json resource version map or use given name
            name = self.versions.get(name, name)
            # return resource uri
            return posixpath.join(self.uri, name)

    def __call__(self, request):
        """Let the instance act as an adapter if needed"""
        return self

    def __repr__(self):
        return '<%s %r at %r>' %(self.__class__.__name__, self.version,
            self.uri)



##############################################################################
#
# Copyright (c) 2006 Zope Foundation and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################
"""Templated Resource Processor

$Id$
"""
__docformat__='restructuredtext'
import re
import zope.interface
import zope.component
from p01.cdn import interfaces

# <COMMAND-BEGIN> <ZRT-COMMAND>: <COMMAND-ARGUMENTS> <COMMAND-END>
NAME = r'[a-zA-Z0-9_-]*'
COMMAND_REGEX = r'%%s(%s):\s*(.*?)\s*%%s' %NAME

TEXTBLOCK=0
COMMAND=1
EXTRCOMMAND=2

class ZRTProcessor(object):
    """A ZRT Processor"""
    zope.interface.implements(interfaces.IZRTProcessor)

    commandStartRegex = r'/\*\s*zrt-'
    commandEndRegex = r'\*/'

    def __init__(self, source, commands=None):
        self.source = source
        if not commands:
            commands = {}
        self.commands = commands
        self._bytecode = None

    def compile(self):
        """See interfaces.IZRTProcessor"""
        bytecode = []
        pos = 0
        # Regular Expression to find commands.
        regex = re.compile(COMMAND_REGEX %(self.commandStartRegex,
                                           self.commandEndRegex))
        nextlinelen = 1
        if '\r\n' in self.source:
            # work around windows: it has a 2 char next line sequence
            nextlinelen = 2

        # Find all commands
        for match in regex.finditer(self.source):
            command, args = match.groups()

            # Add the previous text block and update position
            bytecode.append((TEXTBLOCK, self.source[pos:match.start()]))
            pos = match.end() + nextlinelen

            # Make sure the command exists
            if command not in self.commands:
                cmd = zope.component.queryUtility(
                    interfaces.IZRTCommandFactory, command)

                if cmd is None:
                    raise interfaces.UnknownZRTCommand(command)

                # Add the command
                bytecode.append((EXTRCOMMAND,
                                 (cmd, args, match.start(), match.end())))
            else:
                # Add the command
                bytecode.append(
                    (COMMAND, (command, args, match.start(), match.end())))

        # Add the final textblock
        bytecode.append((TEXTBLOCK, self.source[pos:]))

        self._bytecode = bytecode

    def process(self, context, request):
        """See interfaces.IZRTProcessor"""
        if self._bytecode is None:
            self.compile()

        # List active commands
        active = []
        result = []
        for type, value in self._bytecode:
            # If the type is a command, simply add it to the list of commands
            if type is COMMAND:
                cmd, args, start, end = value
                active.append(
                    self.commands[cmd](args, start, end))
            # If the type is a external command
            elif type is EXTRCOMMAND:
                cmd, args, start, end = value
                active.append(cmd(args, start, end))
            # If we have a textblock, then work on it.
            elif type is TEXTBLOCK:
                # Process any command
                for cmd in active:
                    value = cmd.process(value, context, request)
                    # If the command is no longer available, remove it
                    if not cmd.isAvailable:
                        active.remove(cmd)
                # This block is fully processed now.
                result.append(value)
            else:
                raise ValueError(type)

        return ''.join(result)



##############################################################################
#
# Copyright (c) 2009 Projekt01 GmbH and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################
"""
$Id:$
"""
from __future__ import absolute_import
__docformat__ = "reStructuredText"

import zope.schema
import zope.interface
import zope.component.interfaces
import zope.interface.common.interfaces



class UnknownZRTCommand(ValueError):
    """An unknown ZRT Command was specified"""


class ArgumentError(ValueError):
    """Error while parsing the command arguments."""


class IZRTProcessor(zope.interface.Interface):
    """ZRT Processor"""

    source = zope.schema.Bytes(
        title=u'Source',
        description=u'The source of the expression.')

    def compile():
        """Compile the source to binary form."""

    def process(context, request):
        """Process the source with given context and request.

        Return the result string.
        """


class IZRTCommand(zope.interface.Interface):
    """A ZRT command"""

    isAvailable = zope.schema.Bool(
        title=u'Is Available',
        description=u'Tell whether the command is still available.')

    def process(text, context, request):
        """Process the given text with given context and request.

        Return the result string.
        """

class IZRTCommandFactory(zope.component.interfaces.IFactory):
    """ factory for IZRTCommand """


class IZRTExpression(zope.interface.Interface):
    """An expression to be used in a command."""

    source = zope.schema.Bytes(
        title=u'Source',
        description=u'The source of the expression.')

    context = zope.schema.Object(
        title=u'Context',
        description=u'Context of the template, usually a site.',
        schema=zope.interface.Interface)

    request = zope.schema.Object(
        title=u'Request',
        description=u'Request of the template, usually a site.',
        schema=zope.interface.Interface)


class IZRTInputExpression(IZRTExpression):

    def process(text, outputExpr):
        """Evaluate the expression and update the """


class IZRTOutputExpression(IZRTExpression):

    def process(**kw):
        """Process the expression and return the output string.

        The keyword arguments are additional namespaces.
        """


class IResourceNotFound(zope.interface.common.interfaces.IException):
    """NotFound error for resources

    NOTE: this NotFound allows to register a custom NotFound page whcih is
    able to render a 500 status without to invoke the default NotFound page
    including the html layout.

    NOTE: this interface is not inherited from INotFound or ITraversalException
    because we need to prevent that we will get the default INotFound or
    ITraversalException page.
    """


class IResourceManager(zope.interface.Interface):
    """Content delivery network resource manager

    The resource manager knows the resource output path during extraction.
    The resource manager is also used for generate the right url as an adapter
    adapting the resource layer (request).

    The p01.recipe.cdn package is able to extract the cdn resources. See:
    p01.recipe.cdn for more information.
    """

    rawURI = zope.schema.ASCIILine(
        title=u'Raw resource base uri including fomratting arguments',
        description=u'Raw resource base uri including fomratting arguments',
        default='',
        required=False)

    uri = zope.schema.ASCIILine(
        title=u'Formatted resource base uri',
        description=u'Formatted resource base uri',
        default=None,
        required=False)

    version = zope.schema.ASCIILine(
        title=u'Resource manger version',
        description=u'Resource manger version',
        required=True)

    # optional uri substitution arguments
    namespace = zope.schema.ASCIILine(
        title=u'Resource manager namespace',
        description=u'Resource manager namespace',
        default='',
        required=False)

    skin = zope.schema.ASCIILine(
        title=u'Skin name',
        description=u'Skin name',
        default='',
        required=False)

    site = zope.schema.ASCIILine(
        title=u'Site name',
        description=u'Site name',
        default='',
        required=False)

    useSVNVersions = zope.schema.Bool(
        title=u"Use svn version per file (don't enable on extraction)",
        description=u"Use svn version per file (don't enable on extraction)",
        default=False,
        required=False)

    svnVersionSourcePath = zope.schema.ASCIILine(
        title=u'Generated cdn.json resource version source map path',
        description=u'Generated cdn.json resource version source map path',
        required=False)

    zrtReplaces = zope.schema.Text(
        title=u'Generated zrt-replace.less content',
        description=u'Generated zrt-replace.less content',
        default=u'',
        required=False)

    zrtDirReplaces = zope.schema.Text(
        title=u'Generated zrt-replace.less content (directory first)',
        description=u'Generated zrt-replace.less content (directory first)',
        default=u'',
        required=False)

    svnZRTReplacePath = zope.schema.ASCIILine(
        title=u'Generated zrt-replace.less file path',
        description=u'Generated zrt-replace.less file path',
        required=False)

    versions = zope.schema.Dict(
        title=u'File name (svn) version mapping',
        description=u'File name (svn) version mapping',
        required=False)

    # Note, this output is optional and only required for resource extraction
    # If this output is None, the p01.recipe.cdn recipe option ``output`` is
    # used.
    # See p01.recipe.cdn extract.py for more information
    # Note: if `the recipe output option is used, the namespace will get
    # ignored
    output = zope.schema.ASCIILine(
        title=u'Resource extract output dir path',
        description=u'Resource extract output dir path',
        default=None,
        required=False)

    def getURI(name=None):
        """Build the correct url based on the uri, namespace and version and
        resource name.

        We also, allow to use * as a version manager marker.

        An initial devmode uri could look like:

        http://localhost:8080/++skin++Admin/%(version)s/@@

        If your applicaiton uses site and subsite and each sub site is useing
        another resource manager, the relative subsite path could get used
        as namespace. Then you uri could look like:

        http://localhost:8080/++skin++Admin/%(namespace)s/%(version)s/@@

        And the sub site could be set to something like: "root/subsite"

        A production setup does not require any special uri setup in general.
        It fully depends on your web servers rewrite rule and your extracted
        resource location. You will probably use a sub domain like:

        http://%(namespace)s.foobar.com/%(version)s/@@

        or with an additional namespace:

        http://cdn.foobar.com/%(namespace)s/%(version)s/@@

        or just as minimal as possible

        http://cdn.foobar.com/%(version)s/@@

        """


class ICDNResource(zope.interface.Interface):
    """Content delivery network (cdn) resource.

    This offload resource allows us to configure resources with custom urls.
    This is used if you like to offload static resources form the zope and
    it's fornt end proxy server. Each off load resource contains it's own url
    which can point to another subdomain.

    Note, take care if you use SSL, make sure that you use a wildcard
    certificate which is valid for different subdomains. This is a new kind of
    certificate which only some certificate seller offer. One of such a seller
    is www.godaddy.com. Take care and don't get confused, normaly a wildcard
    certificate is only valid for one domain but for more then one server IP
    address.
    """

    manager = zope.schema.Field(u"IResourceManager instance")

    pkgVersion = zope.schema.TextLine(
        title=u'Package version (only available for resources from eggs)',
        description=u'Package version (only available for resources from eggs)',
        readonly=True,
        required=False)

    def publishTraverse(request, name):
        """Raise NotFound if someone tries to traverse it"""

    def browserDefault(request):
        """Return a callable for processing browser requests."""

    def getZRTReplace(name, version, options):
        """Returns the related zrt-replace directive"""

    def GET():
        """Return the file data for downloading with GET requests."""

    def HEAD():
        """Return proper headers and no content for HEAD requests."""

    def doGET(context):
        """Return the file data for downloading with GET requests."""

    def __call__():
        """Returns the uri for serve as cdn resource"""


class IZRTCDNResource(ICDNResource):
    """CDN zrt resource."""


class II18nCDNResource(ICDNResource):
    """CDN i18n resource."""


class ICDNResourceDirectory(ICDNResource):
    """CDN resource directory."""

    def __getitem__(name):
        """Retruns a resource from the folder

        Raises KeyError if not found.
        """

    def get(name, default=None):
        """Retruns a resource from the folder

        Raises p01.cdn.exceptions.ResourceNotFound if not found.
        """



##############################################################################
#
# Copyright (c) 2009 Projekt01 GmbH and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################
"""Resource NotFound exception page

$Id:$
"""
from __future__ import absolute_import
__docformat__ = "reStructuredText"

from zope.publisher.browser import BrowserPage


class ResourceNotFoundPage(BrowserPage):
    """Resource NotFound page"""

    def __call__(self):
        """Return no content since this renders a not found resource"""
        self.request.response.setStatus(404)
        return u""



##############################################################################
#
# Copyright (c) 2009 Projekt01 GmbH and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################
"""
$Id:$
"""
from __future__ import absolute_import
from zope.interface import implementer
__docformat__ = "reStructuredText"

import zope.interface
import zope.component
import zope.location
from zope.publisher.browser import BrowserView
from zope.publisher.interfaces.browser import IBrowserPublisher

import p01.cdn.exceptions
from p01.cdn import interfaces


def empty():
    return ''


@implementer(IBrowserPublisher)
class Resources(BrowserView):
    """A view that can be traversed further to access browser resources."""


    def publishTraverse(self, request, name):
        """See zope.publisher.interfaces.browser.IBrowserPublisher interface"""
        resource = zope.component.queryAdapter(request, name=name)
        if resource is None:
            raise p01.cdn.exceptions.ResourceNotFound(self, name)

        zope.location.locate(resource, self.context, name)
        return resource

    def browserDefault(self, request):
        """See zope.publisher.interfaces.browser.IBrowserPublisher interface"""
        return empty, ()

    def __getitem__(self, name):
        """A helper method to make this view usable from templates,
        so resources can be acessed in template like context/@@/<resourcename>.
        """
        return self.publishTraverse(self.request, name)

    def __repr__(self):
        return '<%s %r>' % (self.__class__.__name__, self.__name__)


class VersionResources(Resources):
    """Version resource"""

    _version = None

    @property
    def version(self):
        if self._version is None:
            vm = interfaces.IResourceManager(self.request)
            self._version = vm.version
        return self._version

    def publishTraverse(self, request, name):
        """See zope.publisher.interfaces.browser.IBrowserPublisher interface"""
        # first check the version
        if self.version == name:
            resource = Resources(self.context, self.request)
        else:
            resource = zope.component.queryAdapter(request, name=name)
            if resource is None:
                raise p01.cdn.exceptions.ResourceNotFound(self, name)

        zope.location.locate(resource, self.context, name)
        return resource

    def __repr__(self):
        return '<%s %s>' % (self.__class__.__name__, self._version)



##############################################################################
#
# Copyright (c) 2015 Projekt01 GmbH and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################
"""
$Id:$
"""
from __future__ import absolute_import
from zope.interface import implementer
__docformat__ = "reStructuredText"

import zope.interface

from p01.cdn import interfaces


@implementer(interfaces.IResourceNotFound)
class ResourceNotFound(Exception):
    """NotFound error for resources

    NOTE: this NotFound allows to register a custom NotFound page whcih is
    able to render a 500 status without to invoke the default NotFound page
    including the html layout.
    """




##############################################################################
#
# Copyright (c) 2009 Projekt01 GmbH and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################
"""Resource factories and resources

Note; our resource factories defined below, will process the uri based on the
given resource maanger during resource lookup. This is required since we
support the z3c.baseregistry. If you need more speedup, you can define your
own resource factories and use them in the resource zcml directives.

Especialy if you only use one resource manager, you can simply define a
resource manger and override the getResourceManager method in the resource
factories defined below.

$Id:$
"""
from __future__ import absolute_import
from zope.interface import implementer
from builtins import str
from builtins import object
__docformat__ = "reStructuredText"

import os
import time
import pkg_resources
try:
    from email.utils import formatdate
    from email.utils import parsedate_tz
    from email.utils import mktime_tz
except ImportError:
    # python 2.4
    from email.Utils import formatdate
    from email.Utils import parsedate_tz
    from email.Utils import mktime_tz

import zope.interface
import zope.component
import zope.location
import zope.publisher.http
from zope.component import hooks
from zope.contenttype import guess_content_type
from zope.i18n.interfaces import INegotiator
from zope.publisher.interfaces.browser import IBrowserPublisher

import p01.cdn.exceptions
import p01.cdn.processor
import p01.cdn.zrt
from p01.cdn import interfaces

# package version cache
PKG_VERSION_CACHE = {}

# 10 years expiration date
DEFAULT_CACHE_TIMEOUT = 10 * 365 * 24 * 60 * 60

ZRT_REPLACE = u"""/* zrt-replace: "%(prefix)s%(rName)s" tal"string:${context/++resource++%(fName)s}" */
"""

ZRT_DIR_REPLACE = u"""/* zrt-replace: "%(prefix)s%(rName)s" tal"string:${context/++resource++%(dName)s/%(fName)s}" */
"""

def empty():
    """Callable empty string"""
    return ''


###############################################################################
#
# lazy (egg) package version extraction

def getPackageVersion(pkgName):
    """Returns the package version

    Note: resources define in egg packages are not under version control. We
    will use the egg package version. This has the benefit that such resource
    versions and the generated cdn resource url doesn't get changed for the
    same egg package version.

    All files located in your project and versioned in your svn repository
    will use the repository revision at the time they get commited. This has
    a little ugly side effect if you minify such resources. Then this means you
    need to commit a resource before minify them. After minify the minified
    resource will become another version and you need to setup/generate the
    resource version again. This could slown down the release process a bit.
    But the nice part is, you could automat that with p01.build and a buildbot
    server if you like.

    """
    # try to get the resource version. Note: this only works for resources
    # included from egg packages
    v = PKG_VERSION_CACHE.get(pkgName, None)
    if v is None:
        try:
            dist = pkg_resources.require(pkgName)[0]
            v = dist.version
            PKG_VERSION_CACHE[pkgName] = v
        except Exception:
            return None
    return v


###############################################################################
#
# cache control

def setCacheControl(response, seconds=DEFAULT_CACHE_TIMEOUT):
    # 10 years cache timeout by default
    response.setHeader('Cache-Control', 'public,max-age=%s' % seconds)
    t = time.time() + seconds
    response.setHeader('Expires', formatdate(t, usegmt=True))



###############################################################################
#
# file system file wrapper

class File(object):

    def __init__(self, path, name):
        self.path = path
        self.__name__ = name

        f = open(path, 'rb')
        data = f.read()
        f.close()

        self.content_type = guess_content_type(path, data)[0]
        self.lmt = float(os.path.getmtime(path)) or time.time()
        self.lmh = formatdate(self.lmt, usegmt=True)


###############################################################################
#
# cdn resource factories

class ResourceFactoryBase(object):
    """Resource factory base class

    name is the resource.__name__

    rName is the relative resource traversal path dir/sub/resource: This path
    is the relative traversal path from the cdn resource manger to the resource.
    The cdn resoure manager name is not a part of this name. The uri is built
    based on this name prefixed with the cdn manager base uri

    """

    def __init__(self, path, checker, manager, name, pkgName=None):
        self._file = File(path, name)
        self._checker = checker
        self._manager = manager
        # names used for relative path setup
        self._name = name
        # package version used for cdn extration for non svn files
        self._pkgName = pkgName

    def getResourceManager(self, request):
        """Returns the correct resource manager"""
        return zope.component.getAdapter(request, interfaces.IResourceManager,
            name=self._manager)

    def __call__(self, request):
        raise NotImplementedError("Subclass must implement __call__ method")


class ResourceFactory(ResourceFactoryBase):
    """Resource factory"""

    def __call__(self, request):
        manager = self.getResourceManager(request)
        uri = manager.getURI(self._name)
        resource = CDNResource(manager, self._file, request, uri,
            rName=self._name, pkgName=self._pkgName)
        resource.__Security_checker__ = self._checker
        resource.__name__ = self._name
        return resource


class ZRTResourceFactory(ResourceFactoryBase):
    """ZRT Resource factory."""

    def __call__(self, request):
        manager = self.getResourceManager(request)
        uri = manager.getURI(self._name)
        resource = ZRTCDNResource(manager, self._file, request, uri,
            rName=self._name, pkgName=self._pkgName)
        resource.__Security_checker__ = self._checker
        resource.__name__ = self._name
        return resource


class I18nResourceFactory(object):
    """I18n resource factory."""

    def __init__(self, data, checker, manager, name, defaultLanguage,
        pkgName=None,):
        self.__data = data
        self.__uris = []
        self._manager = manager
        self._name = name
        self.__defaultLanguage = defaultLanguage
        self._checker = checker
        # package version used for cdn extration for non svn files
        self._pkgName = pkgName

    def getResourceManager(self, request):
        """Returns the correct resource manager"""
        return zope.component.getAdapter(request, interfaces.IResourceManager,
            name=self._manager)

    def __call__(self, request):
        manager = self.getResourceManager(request)
        resource = I18nCDNResource(manager, self.__data, request,
            self.__defaultLanguage, rName=self._name, pkgName=self._pkgName)
        resource.__Security_checker__ = self._checker
        resource.__name__ = self._name
        return resource


class SubResourceFactory(ResourceFactoryBase):
    """Resource factory"""

    def __init__(self, path, checker, manager, name, rName=None, pkgName=None):
        super(SubResourceFactory, self).__init__(path, checker, manager, name,
            pkgName=pkgName)
        self._rName = rName

    def __call__(self, request):
        manager = self.getResourceManager(request)
        uri = manager.getURI(self._rName)
        resource = CDNResource(manager, self._file, request, uri,
            rName=self._rName, pkgName=self._pkgName)
        resource.__Security_checker__ = self._checker
        resource.__name__ = self._name
        return resource


class SubDirectoryResourceFactory(object):
    """Sub directory resource factory"""

    def __init__(self, data, path, checker, manager, name, rName=None,
        excludeNames=None, pkgName=None):
        self.data = data
        self._path = path
        self._checker = checker
        self._manager = manager
        self._name = name
        self._rName = rName
        if excludeNames is None:
            excludeNames = []
        self._excludeNames = excludeNames
        # package version used for cdn extration for non svn files
        self._pkgName = pkgName

    def getResourceManager(self, request):
        """Returns the correct resource manager"""
        return zope.component.getAdapter(request, interfaces.IResourceManager,
                                         name=self._manager)

    def __call__(self, request):
        manager = self.getResourceManager(request)
        resource = CDNResourceDirectory(manager, self.data, request,
            self._path, self._name, rName=self._rName,
            excludeNames=self._excludeNames, pkgName=self._pkgName)
        resource.__Security_checker__ = self._checker
        resource.__name__ = self._name
        return resource


class DirectoryResourceFactory(object):
    """Directory resource factory"""
    data = {}

    def __init__(self, path, checker, manager, name, excludeNames=None,
        pkgName=None):
        self._path = path
        self._checker = checker
        self._manager = manager
        self._name = name
        if excludeNames is None:
            excludeNames = []
        self._excludeNames = excludeNames
        # package version used for cdn extration for non svn files
        self._pkgName = pkgName
        self.setupData()

    def getStructure(self, path, data, rNameBase):
        """Populate the directory structure with the right factories during
        zcml configuration loading.

        This allows us to read and setup the directory and file factories
        without to define the uri during zcml configuration. The uri will still
        get built based on the resource manager during resource access.

        """
        for name in os.listdir(path):
            if name.startswith('.'):
                continue
            if name in self._excludeNames:
                continue
            data[name] = {}
            # setup relative resource name
            rName = '%s/%s' % (rNameBase, name)
            objPath = os.path.join(path, name)
            if os.path.isdir(objPath):
                rNameNewBase = '%s/%s' % (rNameBase, name)
                subData = self.getStructure(objPath, {}, rNameNewBase)
                data[name]['factory'] = SubDirectoryResourceFactory(
                    subData, path, self._checker, self._manager,
                    name, rName=rName, excludeNames=self._excludeNames,
                    pkgName=self._pkgName)
            else:
                data[name]['factory'] = SubResourceFactory(
                    objPath, self._checker, self._manager, name, rName=rName,
                    pkgName=self._pkgName)
        return data

    def setupData(self):
        """Setup directory structure"""
        # condition
        if not os.path.isdir(self._path):
            raise TypeError(
                "p01.cdnDirectory must be a directory", self._path)
        # get base path for given directory and start producing nested
        # directory and file structure
        rNameBase = self._name
        self.data = self.getStructure(self._path, {}, rNameBase)

    def getResourceManager(self, request):
        """Returns the correct resource manager"""
        return zope.component.getAdapter(request, interfaces.IResourceManager,
            name=self._manager)

    def __call__(self, request):
        manager = self.getResourceManager(request)
        resource = CDNResourceDirectory(
            manager, self.data, request, self._path, self._name,
            rName=self._name, excludeNames=self._excludeNames,
            pkgName=self._pkgName)
        resource.__Security_checker__ = self._checker
        resource.__name__ = self._name
        return resource


###############################################################################
#
# extraction and  publisher

class CDNExtractMixin(object):
    """CDN resource extraction mixin class"""

    rName = None

    def getRelativePath(self, name):
        """Return the relative resource path"""
        # NOTE: by default the name is the resource.rName only the i18n
        # resource is using the lang as name
        return self.rName


@implementer(IBrowserPublisher)
class ResourcePublisher(CDNExtractMixin):
    """Knows how to serv a resource."""


    # 10 years expiration date
    cacheTimeout = DEFAULT_CACHE_TIMEOUT

    def publishTraverse(self, request, name):
        """Raise NotFound if someone tries to traverse it"""
        raise p01.cdn.exceptions.ResourceNotFound(self, name)

    def browserDefault(self, request):
        """Return a callable for processing browser requests."""
        return getattr(self, request.method), ()

    def doGET(self, context):
        """Return the file data for downloading with GET requests."""
        request = self.request
        response = request.response

        setCacheControl(response, self.cacheTimeout)

        # HTTP If-Modified-Since header handling..
        header = request.getHeader('If-Modified-Since', None)
        if header is not None:
            header = header.split(';')[0]
            # Some proxies seem to send invalid date strings for this
            # header. If the date string is not valid, we ignore it
            # rather than raise an error to be generally consistent
            # with common servers such as Apache (which can usually
            # understand the screwy date string as a lucky side effect
            # of the way they parse it).
            try:
                mod_since = int(mktime_tz(parsedate_tz(header)))
            except:
                mod_since = None
            if mod_since is not None:
                if getattr(context, 'lmt', None):
                    last_mod = int(context.lmt)
                else:
                    last_mod = 0
                if last_mod > 0 and last_mod <= mod_since:
                    response.setStatus(304)
                    return ''

        response.setHeader('Content-Type', context.content_type)
        response.setHeader('Last-Modified', context.lmh)

        f = open(context.path, 'rb')
        data = f.read()
        f.close()
        return data

    def GET(self):
        """Return the file data for downloading with GET requests."""
        return self.doGET(self.context)

    def HEAD(self):
        """Return proper headers and no content for HEAD requests."""
        response = self.request.response
        response.setHeader('Content-Type', self.context.content_type)
        response.setHeader('Last-Modified', self.context.lmh)
        setCacheControl(response, self.cacheTimeout)
        return ''

###############################################################################
#
# cdn resources

@implementer(interfaces.ICDNResource)
class CDNResource(ResourcePublisher, zope.location.Location):
    """CDN resource implementation."""


    def __init__(self, manager, context, request, uri, rName=None, pkgName=None):
        self.manager = manager
        self.context = context
        self.path = context.path
        self.request = request
        self.uri = uri
        # cdn extract attrbutes
        self.rName = rName
        # package version used for cdn extration for non svn files
        self.pkgName = pkgName

    @property
    def pkgVersion(self):
        """Lazy package version based on package name"""
        return getPackageVersion(self.pkgName)

    @property
    def rNameDirectory(self):
        if self.rName is not None and '/' in self.rName:
            return self.rName.split('/')[0]
        else:
            # self.__parent__ and rName setup seems to be broken
            return u"BAD/CDNResource/dName/%s" % self.rName

    @property
    def rNameWithoutDirectory(self):
        if self.rName is not None and '/' in self.rName:
            return '/'.join(self.rName.split('/')[1:])
        else:
            # self.__parent__ and rName setup seems to be broken
            return u"BAD/CDNResource/rNameWithoutDirectory/%s" % self.rName

    def getZRTReplace(self, name, version, options):
        """Returns the related zrt-replace directive"""
        prefix = options.zrtPrefixes.get(self.rName)
        if self.rName is not None and '/' in self.rName:
            if prefix is None:
                prefix = options.zrtDirPrefix
            return ZRT_DIR_REPLACE % {
                'prefix': prefix and prefix or '',
                'rName': self.rName,
                'dName': self.rNameDirectory,
                'fName': self.rNameWithoutDirectory,
                }
        else:
            if prefix is None:
                prefix = options.zrtPrefix
            return ZRT_REPLACE % {
                'prefix': prefix and prefix or '',
                'rName': self.rName,
                'fName': self.__name__,
                }

    def __call__(self):
        return self.uri

    def __repr__(self):
        return '<%s %r>' % (self.__class__.__name__, self.__name__)


@implementer(interfaces.IZRTCDNResource)
class ZRTCDNResource(CDNResource):
    """ZRT CDN resource implementation."""


    def _encodeResult(self, data):
        response = self.request.response
        encoding = zope.publisher.http.getCharsetUsingRequest(
            self.request) or 'utf-8'
        content_type = response.getHeader('Content-Type')

        if isinstance(data, str):
            major, minor, params = zope.contenttype.parse.parse(content_type)

            if 'charset' in params:
                encoding = params['charset']

            try:
                data = data.encode(encoding)
            except UnicodeEncodeError:
                # RFC 2616 section 10.4.7 allows us to return an
                # unacceptable encoding instead of 406 Not Acceptable
                # response.
                encoding = 'utf-8'
                data = data.encode(encoding)

            params['charset'] = encoding
            content_type = "%s/%s;" % (major, minor)
            content_type += ";".join(k + "=" + v for k, v in list(params.items()))
            response.setHeader('Content-Type', content_type)

        # set lenght
        response.setHeader('Content-Length', str(len(data)))
        return data

    def GET(self):
        data = super(ZRTCDNResource, self).GET()
        zp = p01.cdn.processor.ZRTProcessor(data, commands={'replace': p01.cdn.zrt.CDNReplace})
        # Note: we run into a ValueError because of using unicode
        # with application/javascript content_type.
        # Use a DirectResult which will bypass the _implicitResult
        # method from zope.publisher.http line 799
        # return zp.process(hooks.getSite(), self.request)

        # bugfix, encode and use DirectResult
        data = zp.process(hooks.getSite(), self.request)
        body = self._encodeResult(data)
        return zope.publisher.http.DirectResult((body,))


@implementer(interfaces.II18nCDNResource)
class I18nCDNResource(ResourcePublisher, zope.location.Location):
    """I18n CDN resource implementation."""


    _locale = None
    __rNameBase = None

    def __init__(self, manager, data, request, defaultLanguage, rName=None,
        pkgName=None):
        self.manager = manager
        self._data = data
        self.request = request
        self.defaultLanguage = defaultLanguage
        self._locale = self.defaultLanguage
        self.rName = rName
        # keep original rName as reference
        self.__rNameBase = rName
        # package version used for cdn extration for non svn files
        self.pkgName = pkgName

    @property
    def pkgVersion(self):
        """Lazy package version based on package name"""
        return getPackageVersion(self.pkgName)

    def getNameForLocale(self, locale=None):
        """Get resource name for locale"""
        if locale is None:
            locale = self.locale
        if not locale or locale not in self._data:
            locale = self.defaultLanguage
        name, ext = os.path.splitext(self.__rNameBase)
        return '%s-%s%s' % (name, self.locale, ext)

    def getAvailableLocales(self):
        """Returns available locales"""
        for locale in list(self._data.keys()):
            yield locale

    def locale():
        """Switch locale and resource name (rName) for extraction"""
        def fget(self):
            return self._locale
        def fset(self, locale):
            if self._locale != locale:
                if locale in self._data:
                    self._locale = locale
                else:
                    self._locale = self.defaultLanguage
                # switch rName for given locale
                self.rName = self.getNameForLocale(self._locale)
        return property(fget, fset)

    locale = locale()
    @property
    def path(self):
        """Return file path for extraction"""
        f = self._data[self.locale]
        return f.path

    def getExtractableResources(self):
        for locale in self.getAvailableLocales():
            # switch locale
            self.locale = locale
            yield self
        # reset to default locale
        self._locale = self.defaultLanguage
        self.rName = self.__rNameBase

    # cdn extract api
    def getRelativePath(self, name):
        """Return the relative resource path"""
        return self.rName

    def getZRTReplace(self, name, version, options):
        """Returns the related zrt-replace directive"""
        prefix = options.zrtPrefixes.get(self.rName)
        if prefix is None:
            prefix = options.zrtPrefix
        return ZRT_REPLACE % {
            'prefix': prefix and prefix or '',
            'rName': self.rName,
            'fName': self.__name__,
            }

    def getURI(self, locale):
        rName = self.getNameForLocale(locale)
        return self.manager.getURI(rName)

    def getPath(self, locale):
        f = self._data[locale]
        return f.path

    def getPaths(self):
        """Used for extract paths, see p01.recipe.cdn/extract.py"""
        for f in list(self._data.values()):
            yield f.path

    def getURIs(self):
        """Used for extract uris, see p01.recipe.cdn/extract.py"""
        for locale in list(self._data.keys()):
            yield self.getURI(locale)

    # serving resource data
    def setUpRequestLocale(self):
        """Setup negotiated locale"""
        langs = list(self._data.keys())
        negotiator = zope.component.getUtility(INegotiator)
        self.locale = negotiator.getLanguage(langs, self.request)

    def GET(self):
        """Return the file data for downloading with GET requests."""
        # first ensure a context
        self.setUpRequestLocale()
        try:
            context = self._data[self.locale]
        except KeyError:
            context = self._data[self.defaultLanguage]
        return self.doGET(context)

    def __call__(self):
        """Returns the uri for serve as cdn resource"""
        self.setUpRequestLocale()
        try:
            return self.getURI(self.locale)
        except KeyError:
            return self.getURI(self.defaultLanguage)

    def __repr__(self):
        return '<%s %r>' % (self.__class__.__name__, self.__name__)


# Directory resource
@implementer(interfaces.ICDNResourceDirectory)
class CDNResourceDirectory(ResourcePublisher, zope.location.Location):


    def __init__(self, manager, data, request, path, name, rName=None,
                 excludeNames=None, pkgName=None):
        self.manager = manager
        self.data = data
        self.request = request
        self.path = path
        self.__name__ = name
        self.rName = rName
        if excludeNames is None:
            excludeNames = []
        self.excludeNames = excludeNames
        # package version used for cdn extration for non svn files
        self.pkgName = pkgName

    @property
    def pkgVersion(self):
        """Lazy package version based on package name"""
        return getPackageVersion(self.pkgName)

    def publishTraverse(self, request, name):
        """See interface IBrowserPublisher"""
        return self.get(name)

    def browserDefault(self, request):
        """See interface IBrowserPublisher"""
        return empty, ()

    @property
    def uri(self):
        return self()

    def __call__(self):
        return self.manager.getURI(self.rName)

    def __getitem__(self, name):
        """Retruns a resource from the folder

        Raises KeyError if not found.
        """
        res = self.get(name, None)
        if res is None:
            raise KeyError(name)
        return res

    def get(self, name, default=None):
        """Retruns a resource from the folder

        Raises p01.cdn.exceptions.ResourceNotFound if not found.
        """
        if name not in self.excludeNames:
            data = self.data.get(name)
            if data is not None:
                factory = data['factory']
                resource = factory(self.request)
                resource.__parent__ = self
                return resource
        # we don't return default, we raise not available
        raise p01.cdn.exceptions.ResourceNotFound(self, name)



##############################################################################
#
# Copyright (c) 2009 Projekt01 GmbH and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################
"""Resource registration



$Id:$
"""
from __future__ import absolute_import
__docformat__ = "reStructuredText"

import os

import zope.interface
import zope.component
import zope.schema
import zope.configuration.fields
import zope.security.zcml
from zope.configuration.exceptions import ConfigurationError
from zope.publisher.interfaces.browser import IBrowserRequest
from zope.security.checker import CheckerPublic
from zope.security.checker import NamesChecker
from zope.security.zcml import Permission
from zope.component.zcml import adapter
from zope.component.zcml import handler

import p01.cdn.layer
from p01.cdn import interfaces
from p01.cdn.resource import File
from p01.cdn.resource import ResourceFactory
from p01.cdn.resource import I18nResourceFactory
from p01.cdn.resource import ZRTResourceFactory
from p01.cdn.resource import DirectoryResourceFactory


allowed_names = ('GET', 'HEAD', 'publishTraverse', 'browserDefault',
                 'request', '__call__', 'uri')


###############################################################################
#
# directive interfaces

class IResourceManagerDirective(zope.interface.Interface):

    """ResourceManager (adapter) directive"""

    factory = zope.configuration.fields.Tokens(
        title=u"Adapter factory/factories",
        description=u"Adapter factory/factories",
        required=True,
        value_type=zope.configuration.fields.GlobalObject()
    )

    provides = zope.configuration.fields.GlobalInterface(
        title=u"Interface the resource manager provides",
        description=u"This attribute specifies the interface the adapter"
        u" instance must provide.",
        required=False,
    )

    layer = zope.configuration.fields.GlobalInterface(
        title=u"The layer the resource should be found in",
        description=u"The layer the resource should be found in",
        required=False
    )

    name = zope.schema.TextLine(
        title=u"Name",
        description=u"Adapters can have names.\n\n"
                    u"This attribute allows you to specify the name for"
                    u" this adapter.",
        required=False,
    )

    permission = Permission(
        title=u"Permission",
        description=u"Permission",
        required=False,
    )

    # versions = zope.configuration.fields.Path(
    #     title=u"Resource name map used for (svn) versioning per resource",
    #     description=u"Resource name map used for (svn) versioning per resource",
    #     required=True
    # )


class IBasicResourceInformation(zope.interface.Interface):

    """Basic resource data."""

    manager = zope.schema.TextLine(
        title=u"Responsible resource manager name",
        description=u"""Responsible resource manager name
        This name is used for register as named adapter lookup. Note, normaly
        you only need to use a named resource manager and the resource manager
        name if you use more then one resource manager and you need to lookup
        resources from a resource manager from an inherited layer.""",
        required=False,
        default=u""
    )

    name = zope.schema.TextLine(
        title=u"The name of the resource",
        description=u"The resource adapter name.",
        required=True
    )

    layer = zope.configuration.fields.GlobalInterface(
        title=u"The layer the resource should be found in",
        description=u"""
        For information on layers, see the documentation for the skin
        directive. Defaults to "default".""",
        required=False
    )

    permission = zope.security.zcml.Permission(
        title=u"The permission needed to access the resource.",
        description=u"""
        If a permission isn't specified, the resource will always be
        accessible.""",
        required=False
    )

    # you can use an optmized resource which is able to skip the resource
    # manager adapter lookup if you need even more speedup. Note, if a
    # z3c.baseregistry is involved, you need to make sure your resource
    # manager will get served from the right site component registry
    factory = zope.configuration.fields.Tokens(
        title=u"Resource factory",
        description=u"Resource factory",
        required=False,
        value_type=zope.configuration.fields.GlobalObject()
    )


class ICDNResourceDirective(IBasicResourceInformation):

    """Defines a browser resource with a explicit url.

    This allows to configure resources with custom urls. This is used if you
    like to cdn static resources form the zope and it's fornt end proxy
    server.
    """

    file = zope.configuration.fields.Path(
        title=u"File",
        description=u"The file containing the resource data.",
        required=False
    )


class IZRTCDNResourceDirective(ICDNResourceDirective):

    """Defines a ZRT browser resource with a explicit url."""


class ICDNI18nResourceDirective(IBasicResourceInformation):

    """Defines a directory containing files.

    The <p01:i18nResource> offers all files in the folder as i18n aware
    resources. This means the resource file must provide usefull locale names.

    Note, only files with the same file ending get uased as resources. And you
    can define exclude names which get not used as resources.
    """

    excludeNames = zope.schema.List(
        title=u"The names which get excluded from registration as resources",
        description=u"""This names get excluded from register as resources""",
        value_type=zope.schema.TextLine(
            title=u"Exclude name",
            description=u"Name to exclude from registration",
            required=True),
        required=False
    )

    directory = zope.configuration.fields.Path(
        title=u"Directory",
        description=u"The directory containing the resource data.",
        required=True
    )

    defaultLanguage = zope.schema.TextLine(
        title=u"Default language",
        description=u"Defines the default language",
        default=u'en',
        required=False
    )

    i18nFactory = zope.configuration.fields.Tokens(
        title=u"I18nResource factory",
        description=u"I18nResource factory",
        required=False,
        value_type=zope.configuration.fields.GlobalObject()
    )


class ICDNResourceDirectoryDirective(IBasicResourceInformation):

    """Defines a directory containing offload resources."""

    directory = zope.configuration.fields.Path(
        title=u"Directory",
        description=u"The directory containing the resource data.",
        required=True
    )

    excludeNames = zope.schema.List(
        title=u"The names which get excluded from traversing as resources",
        description=u"""This names get excluded from traversing as resources""",
        value_type=zope.schema.TextLine(
            title=u"Exclude name",
            description=u"Name to exclude from traversing as resource",
            required=True),
        required=False
    )


###############################################################################
#
# directives

def cdnResourceManager(_context, factory, layer=p01.cdn.layer.ICDNRequest,
    provides=None, name='', permission='zope.Public'):
    """Resource Manager (adapter) directive"""
    if provides is None:
        if len(factory) == 1:
            p = list(zope.interface.implementedBy(factory[0]))
            if len(p) == 1:
                provides = p[0]

        if provides is None:
            provides = interfaces.IResourceManager

    # register the resource manager adapter
    adapter(_context, factory, provides=provides,
            for_=(layer,), permission=permission, name=name, trusted=False,
            locate=False)


# lazy apply resource manager during registration does not work since we
# support the z3c.baseregistry. We need to lookup the right resource manager
# during calling a resource. Otherwise we can't ensure that we get the right
# resource manager
#def registerCDNResource(methodName, *args, **kwargs):
#    """Lazy apply resource manager and register resource"""
#    factory = args[0]
#    layer = args[1][0]
#    name = args[3]
#    sm = zope.component.getSiteManager()
#    mFactory = sm.adapters.lookup1(layer, interfaces.IResourceManager)
#    manager = mFactory(None)
#    factory.setResourceManager(manager)
#    method = getattr(sm, methodName)
#    method(*args, **kwargs)
def registerCDNResource(methodName, *args, **kwargs):
    """Lazy apply resource manager and register resource"""
    sm = zope.component.getSiteManager()
    method = getattr(sm, methodName)
    method(*args, **kwargs)


def cdnResource(_context, name, manager=u'', layer=p01.cdn.layer.ICDNRequest,
    permission='zope.Public', file=None, factory=ResourceFactory):
    """This cdn resource directive allows to register a file as resource"""

    if permission == 'zope.Public':
        permission = CheckerPublic

    rName = name
    pkgName = _context.context.package.__name__

    checker = NamesChecker(allowed_names, permission)
    obj = factory(file, checker, manager, name, pkgName=pkgName)

    _context.action(
        discriminator=('resource', name, IBrowserRequest, layer),
        callable=registerCDNResource,
        args=('registerAdapter', obj, (layer,), interfaces.ICDNResource,
              name, _context.info),
    )


def cdnZRTResource(_context, name, manager=u'', layer=p01.cdn.layer.ICDNRequest,
    permission='zope.Public', file=None, factory=ZRTResourceFactory):
    """This ZRT cdn resource directive allows to register a file as resource"""

    if permission == 'zope.Public':
        permission = CheckerPublic

    pkgName = _context.context.package.__name__

    checker = NamesChecker(allowed_names, permission)
    obj = factory(file, checker, manager, name, pkgName=pkgName)

    _context.action(
        discriminator=('resource', name, IBrowserRequest, layer),
        callable=registerCDNResource,
        args=('registerAdapter', obj, (layer,), interfaces.ICDNResource,
              name, _context.info),
    )


def cdnResourceDirectory(_context, name, manager=u'', directory=None,
                         excludeNames=[], layer=p01.cdn.layer.ICDNRequest,
                         permission='zope.Public',
                         factory=DirectoryResourceFactory):
    """This i18n cdn resource directive allows to register folders as
    resources.

    """
    if not os.path.isdir(directory):
        raise ConfigurationError("Directory %s does not exist" % directory)

    if permission == 'zope.Public':
        permission = CheckerPublic

    pkgName = _context.context.package.__name__

    checker = NamesChecker(allowed_names + ('__getitem__', 'get'), permission)

    obj = factory(directory, checker, manager, name, excludeNames=excludeNames,
        pkgName=pkgName)
    _context.action(
        discriminator=('resource', name, IBrowserRequest, layer),
        callable=registerCDNResource,
        args=('registerAdapter', obj, (layer,), interfaces.ICDNResource,
              name, _context.info),
    )


def cdnI18NResource(_context, name, manager=u'', directory=None,
                    defaultLanguage=u'en', excludeNames=[],
                    layer=p01.cdn.layer.ICDNRequest,
                    permission='zope.Public', i18nFactory=I18nResourceFactory,
                    factory=ResourceFactory):
    """This i18n cdn resource directive allows to register folders as
    resources.

    A resource must be a file name and the folder must contain files including
    this resource name and provide a locale with the file extension. e.g.
    logo-de.gif

    """
    if not os.path.isdir(directory):
        raise zope.configuration.exceptions.ConfigurationError(
            "Directory %s does not exist" % directory)

    # resource name, e.g. logo.gif
    if not '.' in name:
        raise zope.configuration.exceptions.ConfigurationError(
            "Resource %s must be a file name not a directory" % name)
    rName = os.path.splitext(name)[0]

    if permission == 'zope.Public':
        permission = CheckerPublic

    pkgName = _context.context.package.__name__

    checker = NamesChecker(allowed_names, permission)

    # collect relevant resource files
    data = {}
    for fName in os.listdir(directory):
        if fName in excludeNames:
            continue
        # never include (private) .svn folders
        elif fName.startswith('.'):
            continue
        # only include files starting with the same name as the resource name
        elif not fName.startswith(rName):
            raise zope.configuration.exceptions.ConfigurationError(
                "Resource %s must startwith resource name" % (fName, name))
        elif not '-' in fName:
            raise zope.configuration.exceptions.ConfigurationError(
                "Resource %s must contain a locale at the end separated with "
                "- e.g. logo-en.gif, logo-de.gif etc." % (fName, name))
        else:
            # setup resource name e.g. logo-de.gif
            # split file name into plain name and language locale, cut extension
            pName, pExt = os.path.splitext(fName)
            # extract language locale
            locale = pName.replace('%s-' % rName, '')
            # add the file as a tuple to the data dict
            path = os.path.join(directory, fName)
            data[locale] = File(path, fName)

            # register resource for each language used for traversal
            obj = factory(path, checker, manager, fName, pkgName=pkgName)
            _context.action(
                discriminator=('resource', fName, IBrowserRequest, layer),
                callable=registerCDNResource,
                args=('registerAdapter', obj, (layer,),
                      interfaces.ICDNResource, fName, _context.info),
            )

    # register i18n resource factory used for lookup via ++resource++foo
    obj = i18nFactory(data, checker, manager, name, defaultLanguage)
    _context.action(
        discriminator=('resource', name, IBrowserRequest, layer),
        callable=registerCDNResource,
        args=('registerAdapter', obj, (layer,), interfaces.ICDNResource,
              name, _context.info),
    )



##############################################################################
#
# Copyright (c) 2009 Projekt01 GmbH and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################
"""
$Id:$
"""
from __future__ import absolute_import
__docformat__ = "reStructuredText"


import zope.interface


class ICDNRequest(zope.interface.Interface):
    """Request marker interface for make resources (@@) view available.
    
    You don't have to use this view but make sure the VersionResources
    view is available in your layer/skin.
    """
